@@ -1,12 +1,13 @@
 package Data::CSel;
 
-our $DATE = '2016-03-20'; # DATE
-our $VERSION = '0.01'; # VERSION
+our $DATE = '2016-03-22'; # DATE
+our $VERSION = '0.02'; # VERSION
 
 use 5.010001;
 use strict;
 use warnings;
 
+use Code::Includable::Tree::NodeMethods;
 #use List::Util qw(first);
 use Scalar::Util qw(refaddr looks_like_number);
 
@@ -121,6 +122,20 @@ our $RE =
                           $^R->[0][1]{value} = $^R->[1][2] if @{$^R->[1]} > 2;
                           $^R->[0];
                       })
+                  |
+                      \.((?&TYPE_NAME))
+                      (?{
+                          $^R->[1]{type}  = 'class_selector';
+                          $^R->[1]{class} = $^N;
+                          $^R;
+                      })
+                  |
+                      \#(\w+)
+                      (?{
+                          $^R->[1]{type} = 'id_selector';
+                          $^R->[1]{id}   = $^N;
+                          $^R;
+                      })
                   |
                       (?&PSEUDOCLASS) # [[$^R, {}], [$pseudoclass]]
                       (?{
@@ -158,7 +173,7 @@ our $RE =
 
                   (
                       \s*(?:=~|!~)\s* |
-                      \s*(?:!=|>=?|<=?|==?)\s* |
+                      \s*(?:!=|<>|>=?|<=?|==?)\s* |
                       \s+(?:eq|ne|lt|gt|le|ge)\s+ |
                       \s+(?:isnt|is)\s+
                   )
@@ -270,130 +285,6 @@ sub parse_csel {
     return undef;
 }
 
-sub _node_children {
-    my $node = shift;
-    my @c = $node->children;
-    @c = @{$c[0]} if @c==1 && ref($c[0]) eq 'ARRAY';
-    @c;
-}
-
-sub _node_descendants {
-    my $node = shift;
-    my @c = _node_children($node);
-    (@c, map { _node_descendants($_) } _node_children($node));
-}
-
-sub _node_is_first_child {
-    my $node = shift;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my @c = _node_children($parent);
-    @c && refaddr($node) == refaddr($c[0]);
-}
-
-sub _node_is_last_child {
-    my $node = shift;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my @c = _node_children($parent);
-    @c && refaddr($node) == refaddr($c[-1]);
-}
-
-sub _node_is_only_child {
-    my $node = shift;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my @c = _node_children($parent);
-    @c==1;# && refaddr($node) == refaddr($c[0]);
-}
-
-sub _node_is_nth_child {
-    my ($node, $n) = @_;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my @c = _node_children($parent);
-    @c >= $n && refaddr($node) == refaddr($c[$n-1]);
-}
-
-sub _node_is_nth_last_child {
-    my ($node, $n) = @_;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my @c = _node_children($parent);
-    @c >= $n && refaddr($node) == refaddr($c[-$n]);
-}
-
-sub _node_is_first_of_type {
-    my $node = shift;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my $type = ref($node);
-    my @c = grep { $_->isa($type) } _node_children($parent);
-    @c && refaddr($node) == refaddr($c[0]);
-}
-
-sub _node_is_last_of_type {
-    my $node = shift;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my $type = ref($node);
-    my @c = grep { $_->isa($type) } _node_children($parent);
-    @c && refaddr($node) == refaddr($c[-1]);
-}
-
-sub _node_is_only_of_type {
-    my $node = shift;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my $type = ref($node);
-    my @c = grep { $_->isa($type) } _node_children($parent);
-    @c == 1; # && refaddr($node) == refaddr($c[0]);
-}
-
-sub _node_is_nth_of_type {
-    my ($node, $n) = @_;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my $type = ref($node);
-    my @c = grep { $_->isa($type) } _node_children($parent);
-    @c >= $n && refaddr($node) == refaddr($c[$n-1]);
-}
-
-sub _node_is_nth_last_of_type {
-    my ($node, $n) = @_;
-    my $parent = $node->parent;
-    return 0 unless $parent;
-    my $type = ref($node);
-    my @c = grep { $_->isa($type) } _node_children($parent);
-    @c >= $n && refaddr($node) == refaddr($c[-$n]);
-}
-
-sub _node_little_siblings {
-    my $node = shift;
-    my $parent = $node->parent or return ();
-    my $refaddr = refaddr($node);
-    my @children = _node_children($parent);
-    for my $i (0..$#children-1) {
-        if (refaddr($children[$i]) == $refaddr) {
-            return @children[$i+1 .. $#children];
-        }
-    }
-    ();
-}
-
-sub _node_adjacent_little_sibling {
-    my $node = shift;
-    my $parent = $node->parent or return undef;
-    my $refaddr = refaddr($node);
-    my @children = _node_children($parent);
-    for my $i (0..$#children-1) {
-        if (refaddr($children[$i]) == $refaddr) {
-            return $children[$i+1];
-        }
-    }
-    undef;
-}
-
 sub _uniq_objects {
     my @uniq;
     my %mem;
@@ -412,14 +303,14 @@ sub _simpsel {
 
     my @res;
     if ($is_recursive) {
-        @res = (@nodes, map {_node_descendants($_)} @nodes);
+        @res = (@nodes, map {Code::Includable::Tree::NodeMethods::descendants($_)} @nodes);
     } else {
         @res = @nodes;
     }
     #say "D:   intermediate result (after walk): [".join(",",map {$_->{id}} @res)."]";
 
     unless ($simpsel->{type} eq '*') {
-        @res = grep { $_->isa($simpsel->{type}) } @res;
+        @res = grep { ref($_) eq $simpsel->{type} } @res;
     }
 
     @res = _uniq_objects(@res);
@@ -428,15 +319,19 @@ sub _simpsel {
     for my $f (@{ $simpsel->{filters} // [] }) {
         last unless @res;
 
-        if (defined (my $attr = $f->{attr})) {
+        my $type = $f->{type};
+
+        if ($type eq 'attr_selector') {
+
+            my $attr = $f->{attr};
             my $op  = $f->{op};
             my $opv = $f->{value};
 
             my @newres;
           ITEM:
             for my $o (@res) {
-                next ITEM unless $o->can($f->{attr}) ||
-                    $o->can('AUTOLOAD');
+                next ITEM unless $o->can($f->{attr});
+                goto PASS unless $op;
 
                 my $val = $o->$attr;
                 if ($op eq '=' || $op eq '==') {
@@ -447,7 +342,7 @@ sub _simpsel {
                     }
                 } elsif ($op eq 'eq') {
                     next ITEM unless $val eq $opv;
-                } elsif ($op eq '!=') {
+                } elsif ($op eq '!=' || $op eq '<>') {
                     if (looks_like_number($opv)) {
                         next ITEM unless $val != $opv;
                     } else {
@@ -467,7 +362,7 @@ sub _simpsel {
                     if (looks_like_number($opv)) {
                         next ITEM unless $val >=  $opv;
                     } else {
-                        next ITEM unless $val gt $opv;
+                        next ITEM unless $val ge $opv;
                     }
                 } elsif ($op eq 'ge') {
                     next ITEM unless $val ge $opv;
@@ -511,38 +406,57 @@ sub _simpsel {
                     die "BUG: Unsupported operator '$op' in attr_selector";
                 }
 
+              PASS:
                 # pass all attribute filters, add to new result
                 #say "D:    adding to result: ".$o->{id};
                 push @newres, $o;
             } # for each item
             @res = @newres;
-        } # attr filter
 
-        if (defined(my $pc = $f->{pseudoclass})) {
+        } elsif ($type eq 'class_selector') {
+
+            my $class = $f->{class};
+            @res = grep { $_->isa($class) } @res;
+
+        } elsif ($type eq 'id_selector') {
+
+            my $method = $opts->{id_method} // 'id';
+            my $id     = $f->{id};
+
+            @res = grep { $_->can($method) && $_->$method eq $id } @res;
+
+        } elsif ($type eq 'pseudoclass') {
+
+            my $pc = $f->{pseudoclass};
+
             if ($pc eq 'first') {
                 @res = ($res[0]);
             } elsif ($pc eq 'last') {
                 @res = ($res[-1]);
             } elsif ($pc eq 'first-child') {
-                @res = grep { _node_is_first_child($_) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_first_child($_) } @res;
             } elsif ($pc eq 'last-child') {
-                @res = grep { _node_is_last_child($_) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_last_child($_) } @res;
             } elsif ($pc eq 'only-child') {
-                @res = grep { _node_is_only_child($_) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_only_child($_) } @res;
             } elsif ($pc eq 'nth-child') {
-                @res = grep { _node_is_nth_child($_, $f->{args}[0]) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_nth_child($_, $f->{args}[0]) } @res;
             } elsif ($pc eq 'nth-last-child') {
-                @res = grep { _node_is_nth_last_child($_, $f->{args}[0]) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_nth_last_child($_, $f->{args}[0]) } @res;
             } elsif ($pc eq 'first-of-type') {
-                @res = grep { _node_is_first_of_type($_) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_first_child_of_type($_) } @res;
             } elsif ($pc eq 'last-of-type') {
-                @res = grep { _node_is_last_of_type($_) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_last_child_of_type($_) } @res;
             } elsif ($pc eq 'only-of-type') {
-                @res = grep { _node_is_only_of_type($_) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_only_child_of_type($_) } @res;
             } elsif ($pc eq 'nth-of-type') {
-                @res = grep { _node_is_nth_of_type($_, $f->{args}[0]) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_nth_child_of_type($_, $f->{args}[0]) } @res;
             } elsif ($pc eq 'nth-last-of-type') {
-                @res = grep { _node_is_nth_last_of_type($_, $f->{args}[0]) } @res;
+                @res = grep { Code::Includable::Tree::NodeMethods::is_nth_last_child_of_type($_, $f->{args}[0]) } @res;
+            } elsif ($pc eq 'root') {
+                @res = grep { !$_->parent } @res;
+            } elsif ($pc eq 'empty') {
+                @res = grep { my @c = Code::Includable::Tree::NodeMethods::_children_as_list($_); !@c } @res;
             } elsif ($pc eq 'has') {
                 @res = grep { csel($opts, $f->{args}[0], $_) }
                     _uniq_objects(
@@ -557,7 +471,8 @@ sub _simpsel {
             } else {
                 die "Unsupported pseudo-class '$pc'";
             }
-        } # pseudoclass filter
+
+        }
 
         #say "D:   intermediate result (after filter): [".join(",",map {$_->{id}} @res)."]";
     } # for each filter
@@ -581,17 +496,17 @@ sub _sel {
             last unless @res;
             if ($combinator->{combinator} eq ' ') { # descendant
                 @res = _simpsel($opts, $simpsel, 1,
-                                map { _node_children($_) } @res);
+                                map { Code::Includable::Tree::NodeMethods::_children_as_list($_) } @res);
             } elsif ($combinator->{combinator} eq '>') { # child
                 @res = _simpsel($opts, $simpsel, 0,
-                                map { _node_children($_) } @res);
+                                map { Code::Includable::Tree::NodeMethods::_children_as_list($_) } @res);
             } elsif ($combinator->{combinator} eq '~') { # sibling
                 @res = _simpsel($opts, $simpsel, 0,
-                                map { _node_little_siblings($_) } @res);
+                                map { Code::Includable::Tree::NodeMethods::next_siblings($_) } @res);
             } elsif ($combinator->{combinator} eq '+') { # adjacent sibling
                 @res = _simpsel($opts, $simpsel, 0,
                                 grep {defined}
-                                    map { _node_adjacent_little_sibling($_) }
+                                    map { Code::Includable::Tree::NodeMethods::next_sibling($_) }
                                     @res);
             } else {
                 die "BUG: Unknown combinator '$combinator->{combinator}'";
@@ -639,13 +554,13 @@ Data::CSel - Select tree node objects using CSS Selector-like syntax
 
 =head1 VERSION
 
-This document describes version 0.01 of Data::CSel (from Perl distribution Data-CSel), released on 2016-03-20.
+This document describes version 0.02 of Data::CSel (from Perl distribution Data-CSel), released on 2016-03-22.
 
 =head1 SYNOPSIS
 
  use Data::CSel qw(csel);
 
- my @cells = csel("Table[name=~/data/i] TCell[value isnt '']:first", $tree);
+ my @cells = csel("Table[name=~/data/i] TCell[value != '']:first", $tree);
 
  # ditto, but wrap result using a Data::CSel::Selection
  my $res = csel({wrap=>1}, "Table ...", $tree);
@@ -677,10 +592,12 @@ F element descendant of an E element. C<< E > F >> means F element child of E
 element. C<E ~ F> means F element preceded by an E element. C<E + F> means F
 element immediately preceded by an E element.
 
-A I<simple selector> is either a type selector or universal selector followed
-immediately by zero or more attribute selectors or pseudo-classes, in any order.
-Type or universal selector is optional if there are at least one attribute
-selector or pseudo-class.
+A I<simple selector> is either a L<type selector|/"Type selector"> or
+L<universal selector|/"Universal selector"> followed immediately by zero or more
+L<attribute selectors|/"Attribute selector"> or L<Class selector|"/Class
+selector"> or L<ID selector|/"ID selector"> or
+L<pseudo-classes|"/Pseudo-class">, in any order. Type or universal selector is
+optional if there is at least one attribute selector or pseudo-class.
 
 =head2 Type selector
 
@@ -690,7 +607,8 @@ Example:
 
  My::Class
 
-will match any C<My::Class> object.
+will match any C<My::Class> object. Subclasses of C<My::Class> will I<not> be
+matched, use L<class selector|"/Class selector"> for that.
 
 =head2 Universal selector
 
@@ -713,19 +631,18 @@ The syntax is:
 C<[ATTR]> means to only select objects that have an attribute named C<ATTR>, for
 example:
 
- Any[length]
+ [length]
 
-means to select object of type (C<isa()>) C<Any> that responds to (C<can()>)
-C<length()>.
+means to select objects that respond to (C<can()>) C<length()>.
 
 Note: to select objects that do not have a specified attribute, you can use the
 C<:not> pseudo-class (see L</"Pseudo-class">), for example:
 
- Any:not([length])
+ :not('[length]')
 
-C<[ATTR]> means to only select objects that have an attribute named C<ATTR> that
-has value that matches the expression specified by operator C<OP> and operand
-C<LITERAL>.
+C<[ATTR OP LITERAL]> means to only select objects that have an attribute named
+C<ATTR> that has value that matches the expression specified by operator C<OP>
+and operand C<LITERAL>.
 
 =head3 Literal
 
@@ -811,13 +728,14 @@ Example:
 selects all C<Table> objects that have C<title()> with the value not equal to
 C<"TOC">.
 
-=item * C<!=>
+=item * C<!=> (or C<< <> >>)
 
 Numerical inequality using Perl's C<!=> operator.
 
 Example:
 
  TableCell[length != 3]
+ TableCell[length <> 3]
 
 selects all C<TableCell> objects that have C<length()> with the value not equal
 to 3.
@@ -997,14 +915,44 @@ will select all Person objects where age is defined.
 
 =back
 
+=head2 Class selector
+
+A I<class selector> is a C<.> (dot) followed by Perl class/package name.
+
+ .CLASSNAME
+
+It selects all objects that C<isa()> a certain class. The difference with L<type
+selector|/"Type selector"> is that inheritance is observed. So:
+
+ .My::Class
+
+will match instances of C<My::Class> as well as subclasses of it.
+
+=head2 ID selector
+
+An I<ID selector> is a C<#> (hash) followed by an identifier:
+
+ #ID
+
+It is a special/shortcut form of attribute selector where the attribute is
+C<id> and the operator is C<=>:
+
+ [id = ID]
+
+The L<csel()/"FUNCTIONS"> function allows you to configure which attribute to
+use as the ID attribute, the default is C<id>.
+
 =head2 Pseudo-class
 
-A I<pseudo-class> filters objects based on some criteria, in the form of:
+A I<pseudo-class> is C<:> (colon) followed by pseudo-class name (a
+dash-separated word list), and optionally a list of arguments enclosed in
+parentheses.
 
- :NAME
- :NAME(ARG, ...)
+ :PSEUDOCLASSNAME
+ :PSEUDOCLASSNAME(ARG, ...)
 
-Supported pseudo-classes include:
+It filters result set based on some criteria. Currently supported pseudo-classes
+include:
 
 =over
 
@@ -1030,37 +978,65 @@ selects the last person whose name starts with the letter C<A>.
 
 =item * C<:first-child>
 
-Select only object that is the first child of its parent.
+Select only objects that are the first child of their parent.
 
 =item * C<:last-child>
 
-Select only object that is the last child of its parent.
+Select only objects that are the last child of their parent.
 
 =item * C<:only-child>
 
-Select only object that is the only child of its parent.
+Select only objects that is the only child of their parent.
 
 =item * C<:nth-child(n)>
 
-Select only object that is the I<n>th child of its parent.
+Select only objects that are the I<n>th child of their parent.
 
 =item * C<:nth-last-child(n)>
 
-Select only object that is the I<n>th last child of its parent.
+Select only objects that are the I<n>th last child of their parent.
 
 =item * C<:first-of-type>
 
+Select only objects that are the first child of their parent of their type. So
+if a parent's children is:
+
+ id1(type=T1) id2(T2) id3(T2)
+
+then both C<id1> and C<id2> are first children of their respective types.
+
 =item * C<:last-of-type>
 
+Select only objects that are the last child of their parent of their type.
+
 =item * C<:only-of-type>
 
+Select only objects that are the only child of their parent of their type.
+
 =item * C<:nth-of-type(n)>
 
+Select only objects that are the I<n>th child of their parent of their type.
+
 =item * C<:nth-last-of-type(n)>
 
-=item * C<:not(s)>
+Select only objects that are the I<n>th last child of their parent of their
+type.
 
-=item * C<:has(s)>
+=item * C<:root>
+
+Select only root node(s).
+
+=item * C<:empty>
+
+Select only leaf node(s).
+
+=item * C<:not(S)>
+
+Select all objects not matching selector C<S>.
+
+=item * C<:has(S)>
+
+Select all objects that have a descendant matching selector C<S>.
 
 =back
 
@@ -1071,15 +1047,6 @@ Select only object that is the I<n>th last child of its parent.
 Since Perl package names are separated by C<::>, CSel allows it in type
 selector.
 
-=head3 No equivalent for CSS class and ID selectors
-
-I.e.:
-
- E.class
- E#id
-
-They are not used in CSel.
-
 =head3 Syntax of attribute selector is a bit different
 
 In CSel, the syntax of attribute selector is made simpler and more regular.
@@ -1109,7 +1076,9 @@ literal and must be quoted.
 
 =head3 There is no concept of CSS namespaces
 
-But Perl package names are already hierarchical.
+CSS namespaces are used when there are foreign elements (e.g. SVG in addition to
+HTML) and one wants to use the same stylesheet for both. There is no need for
+something like this CSel, as we deal with only Perl objects.
 
 =head1 FUNCTIONS
 
@@ -1125,7 +1094,8 @@ A tree node object is any regular Perl object satisfying the following criteria:
 1) it supports a C<parent> method which should return a single parent node
 object, or undef if object is the root node); 2) it supports a C<children>
 method which should return a list (or an arrayref) of children node objects
-(where the list/array will be empty for a leaf node).
+(where the list/array will be empty for a leaf node). Note: you can use
+L<Role::TinyCommons::Tree::Node> to enforce this requirement.
 
 Known options:
 
